"""Module containing attribute filter occurrence data wrangler."""
from lmpy.data_wrangling.occurrence.base import _OccurrenceDataWrangler


# .....................................................................................
class AttributeFilterWrangler(_OccurrenceDataWrangler):
    """Filters points according to a function."""
    name = 'AttributeFilterWrangler'
    version = '1.0'

    # .......................
    def __init__(self, attribute_name, filter_func, **params):
        """Constructor for AttributeModifierWrangler class.

        Args:
            attribute_name (str): The name of the attribute to modify.
            filter_func (Method): A function to be used for the pass condition.
            **params (dict): Extra parameters to be sent to the base class.
        """
        self.attribute_name = attribute_name
        _OccurrenceDataWrangler.__init__(self, **params)
        self._get_pass_condition_func(filter_func)

    # .......................
    def _get_pass_condition_func(self, filter_func):
        """Get the pass condition function.

        Args:
            filter_func (Method or dict): A method to use for pass condition.
        """
        if isinstance(filter_func, dict):
            if 'for-each' in filter_func.keys():
                condition = filter_func['for-each']['condition']
                if condition.lower() == 'not-in':
                    bad_values = filter_func['for-each']['values']

                    def _each_value_not_in(point):
                        return all(
                            [
                                val not in bad_values
                                for val in point.get_attribute(self.attribute_name)
                            ]
                        )

                    self._pass_condition = _each_value_not_in
        else:
            self._pass_condition = filter_func



"""Module containing occurrence data wranglers for filtering points."""
from lmpy.data_wrangling.occurrence.base import _OccurrenceDataWrangler


# .....................................................................................
class MinimumPointsFilter(_OccurrenceDataWrangler):
    """Filter points if there are less than the specified number."""
    name = 'MinimumPointsWrangler'
    version = '1.0'

    # .......................
    def __init__(self, minimum_count, **params):
        """Get an occurrence data wrangler for minimum points.

        Args:
            minimum_count (int): The minimum number of points in order to keep all.
            **params (dict): Extra parameters to be sent to the base class.
        """
        self.minimum_count = minimum_count
        _OccurrenceDataWrangler.__init__(self, **params)

    # .......................
    def _get_all_pass_condition(self):
        def all_pass_condition(point):
            return True
        return all_pass_condition

    # .......................
    def _get_all_fail_condition(self):
        def all_fail_condition(point):
            return False
        return all_fail_condition

    # .......................
    def wrangle_points(self, points):
        """Wrangle points.  Return all or none based on how many are present.

        Args:
            points (list of Point): A list of points to assess.

        Returns:
            list of Point: Either an empty list or all of the points passed in.
        """
        if len(points) >= self.minimum_count:
            # Enough points, so all pass
            self._pass_condition = self._get_all_pass_condition()
        else:
            # Not enough points to all fail
            self._pass_condition = self._get_all_fail_condition()

        # Call parent class wrangle_points function
        return _OccurrenceDataWrangler.wrangle_points(self, points)



"""Module containing a tool for parsing a configuration file for argparse."""
import json


# .....................................................................................
def _process_arguments(parser, config_arg=None):
    """Process arguments including filling in those provided by configuration file.

    Args:
        parser (argparse.ArgumentParser): An argparse.ArgumentParser with parameters.
        config_arg (str): If provided, try to read configuration file for additional
            arguments.

    Returns:
        argparse.Namespace: An augmented Namespace with any parameters specified in a
            configuration file.
    """
    args = parser.parse_args()

    if config_arg is not None and hasattr(args, config_arg):
        config_filename = getattr(args, config_arg)
        if config_filename is not None:
            with open(config_filename, mode='rt') as in_json:
                config = json.load(in_json)
                for k in config.keys():
                    tmp = getattr(args, k)
                    if tmp is None:
                        setattr(args, k, config[k])
                    elif isinstance(tmp, list):
                        tmp.extend(config[k])
                        setattr(args, k, tmp)
    return args



"""Occurrence data wrangling package."""

from . import accepted_name_wrangler
from . import attribute_filter_wrangler
from . import attribute_modifier_wrangler
from . import base
from . import bounding_box_wrangler
from . import common_format_wrangler
from . import coordinate_conversion_wrangler
from . import decimal_precision_wrangler
from . import disjoint_geometries_wrangler
from . import intersect_geometries_wrangler
from . import minimum_points_wrangler
from . import spatial_index_wrangler
from . import unique_localities_wrangler

__all__ = [
    'accepted_name_wrangler',
    'attribute_filter_wrangler',
    'attribute_modifier_wrangler',
    'base',
    'bounding_box_wrangler',
    'common_format_wrangler',
    'coordinate_conversion_wrangler',
    'decimal_precision_wrangler',
    'disjoint_geometries_wrangler',
    'intersect_geometries_wrangler',
    'minimum_points_wrangler',
    'spatial_index_wrangler',
    'unique_localities_wrangler',
]



"""Data wrangling package."""
from . import common
from . import matrix
from . import occurrence
from . import tree

__all__ = []
__all__.extend(common.__all__)
__all__.extend(matrix.__all__)
__all__.extend(occurrence.__all__)
__all__.extend(tree.__all__)



"""Split occurrence data files into groups for processing."""
import argparse
import json

from lmpy.data_preparation.occurrence_splitter import (
    DEFAULT_MAX_WRITERS,
    get_writer_key_from_fields_func,
    get_writer_filename_func,
    OccurrenceSplitter,
)
from lmpy.data_wrangling.factory import WranglerFactory
from lmpy.point import PointCsvReader, PointDwcaReader
from lmpy.tools._config_parser import _process_arguments


# .....................................................................................
DESCRIPTION = '''\
Group and split occurrence data from one or more sources so that like-records (ex. \
species) can be processed together.'''


# .....................................................................................
def cli():
    """Command-line interface for splitting occurrence datasets."""
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        '--config_file',
        type=str,
        help='Configuration file containing script arguments.'
    )
    parser.add_argument(
        '-m',
        '--max_open_writers',
        type=int,
        default=DEFAULT_MAX_WRITERS,
        choices=range(1, 500),
        metavar='[1 - 500]',
        help=(
            'The maximum number of data writers to have open at once. '
            'Too many open files can cause errors. '
            'Default: {}.'.format(DEFAULT_MAX_WRITERS)
        ),
    )
    parser.add_argument(
        '-k',
        '--key_field',
        action='append',
        type=str,
        help=(
            'A field to use to determine writer key.  Multiple values can be used to '
            'utilize multiple fields.'
        ),
    )
    parser.add_argument(
        '-of',
        '--out_field',
        action='append',
        type=str,
        help=(
            'Include this field in the outputs.  If not provided, all fields from the '
            'first point ready for output will be used.'
        ),
    )
    parser.add_argument(
        '--dwca',
        action='append',
        nargs=2,
        help='A Darwin-Core Archive to process and associated wrangler configuration.',
    )
    parser.add_argument(
        '--csv',
        action='append',
        nargs=5,
        help=(
            'A CSV file to process, an associated wrangler configuration file, '
            'a species header key, an x header key, and a y header key.'
        ),
    )

    parser.add_argument(
        'out_dir', type=str, help='Directory where the output data should be written.'
    )
    args = _process_arguments(parser, 'config_file')

    # Establish functions for getting writer key and filename
    writer_key_func = get_writer_key_from_fields_func(*tuple(args.key_field))
    writer_filename_func = get_writer_filename_func(args.out_dir)

    # Determine fields to write
    write_fields = None
    if args.out_field is not None:
        write_fields = args.out_field

    # Wrangler Factory
    wrangler_factory = WranglerFactory()

    # Initialize processor
    with OccurrenceSplitter(
        writer_key_func,
        writer_filename_func,
        write_fields=write_fields,
        max_writers=args.max_open_writers,
    ) as occurrence_processor:
        # For each dwca file
        if args.dwca:
            for dwca_fn, wranglers_fn in args.dwca:
                reader = PointDwcaReader(dwca_fn)
                with open(wranglers_fn, mode='rt') as in_json:
                    wranglers = wrangler_factory.get_wranglers(json.load(in_json))
                occurrence_processor.process_reader(reader, wranglers)
        if args.csv:
            # For each csv file
            for csv_fn, wranglers_fn, sp_key, x_key, y_key in args.csv:
                reader = PointCsvReader(csv_fn, sp_key, x_key, y_key)
                with open(wranglers_fn, mode='rt') as in_json:
                    wranglers = wrangler_factory.get_wranglers(json.load(in_json))
                occurrence_processor.process_reader(reader, wranglers)


# .....................................................................................
__all__ = ['cli']


# .....................................................................................
if __name__ == '__main__':  # pragma: no cover
    cli()



"""Module containing occurrence data wranglers for filtering points."""
from lmpy.data_wrangling.occurrence.base import _OccurrenceDataWrangler
from lmpy.spatial import SpatialIndex


# .....................................................................................
class SpatialIndexFilter(_OccurrenceDataWrangler):
    """Spatial Index Filter Occurrence Data Wrangler."""
    name = 'SpatialIndexFilter'
    version = '1.0'

    # .......................
    def __init__(
        self,
        spatial_index,
        intersections_map,
        check_hit_func,
        **params,
    ):
        """A data wrangler to filter points using a spatial index.

        Args:
            spatial_index (SpatialIndex): A SpatialIndex object that can be searched.
            intersections_map (dict): A dictionary of species name keys and
                corresponding valid intersection values.
            check_hit_func (Method): A function that takes two arguments
                (search hit, valid intersections for a species) and returns a boolean
                indication if the hit should be counted.
            **params (dict): Extra parameters to be sent to the base class.
        """
        self.spatial_index = spatial_index
        if not isinstance(self.spatial_index, SpatialIndex):
            self.spatial_index = SpatialIndex(spatial_index)
        self.intersections_map = intersections_map
        self.check_hit_func = check_hit_func

        _OccurrenceDataWrangler.__init__(self, **params)

    # .......................
    def _pass_condition(self, point):
        """Assess a point to see if it passes the spatial index filter.

        Args:
            point (Point): A point object to assess.

        Returns:
            bool: Indication if the point passess the assessment.
        """
        if any(
            [
                point.species_name not in self.intersections_map.keys(),
                len(self.intersections_map[point.species_name]) == 0
            ]
        ):
            return True
        for hit in self.spatial_index.search(point.x, point.y).values():
            if self.check_hit_func(hit, self.intersections_map[point.species_name]):
                return True
        return False



"""Module containing data wranglers for subsetting a tree to match a matrix."""
from lmpy.matrix import Matrix

from lmpy.data_wrangling.tree.base import _TreeDataWrangler


# .....................................................................................
class MatchMatrixTreeWrangler(_TreeDataWrangler):
    """Subsets a tree to match the species in the matrix."""
    name = 'MatchMatrixTreeWrangler'
    version = '1.0'

    # .......................
    def __init__(
        self,
        matrix,
        species_axis,
        **params
    ):
        """Constructor for SubsetTreeWrangler class.

        Args:
            matrix (Matrix): A matrix to get taxon names to match.
            species_axis (int): The matrix axis with taxon names.
            **params (dict): Keyword parameters to pass to _MatrixDataWrangler.
        """
        _TreeDataWrangler.__init__(self, **params)
        if isinstance(matrix, str):
            matrix = Matrix.load(matrix)
        self.keep_taxon_names = matrix.get_headers(axis=str(species_axis))

    # .......................
    def wrangle_tree(self, tree):
        """Wrangle a tree.

        Args:
            tree (TreeWrapper): A tree to wrangle.

        Returns:
            TreeWrapper: The subsetted tree.
        """
        original_taxa_count = len(tree.taxon_namespace)
        tree.retain_taxa_with_labels(self.keep_taxon_names)
        tree.purge_taxon_namespace()
        num_purged = original_taxa_count - len(tree.taxon_namespace)
        self._purged += num_purged
        return tree



"""Module containing data wranglers for subsetting and reordering matrix slices."""
from lmpy.data_wrangling.matrix.base import _MatrixDataWrangler


# .....................................................................................
class SubsetReorderSlicesWrangler(_MatrixDataWrangler):
    """Subsets and / or reorders matrix slices."""
    name = 'SubsetReorderSlicesWrangler'
    version = '1.0'

    # .......................
    def __init__(
        self,
        axes,
        **params
    ):
        """Constructor for PurgeEmptySlicesWrangler class.

        Args:
            axes (dict): A dictionary of axis keys (int or str of int) and list of
                values of those headers to keep.
            **params (dict): Keyword parameters to pass to _MatrixDataWrangler.
        """
        _MatrixDataWrangler.__init__(self, **params)
        # Convert axis keys to integers
        self.slice_axes = {int(k): axes[k] for k in axes.keys()}
        self.report['changes'] = {}

    # .......................
    def wrangle_matrix(self, matrix):
        """Wrangle a matrix.

        Args:
            matrix (Matrix): A matrix to wrangle.

        Returns:
            Matrix: The matrix with empty slices purged.
        """
        slices = []
        for axis in range(matrix.ndim):
            if axis in self.slice_axes.keys():
                axis_slice = []
                axis_headers = matrix.get_headers(axis=str(axis))
                for idx in self.slice_axes[axis]:
                    # If the header is in the axis headers, append the index
                    if idx in axis_headers:
                        axis_slice.append(axis_headers.index(idx))

                if str(axis) not in self.report['changes'].keys():
                    self.report['changes'][str(axis)] = {'purged': 0}
                self.report[
                    'changes'
                ][str(axis)]['purged'] += (len(axis_headers) - len(axis_slice))
            else:
                axis_slice = list(range(matrix.shape[axis]))

            slices.append(axis_slice)

        return matrix.slice(*slices)



"""Module containing data wranglers for matching a matrix to a tree."""
from lmpy.tree import TreeWrapper

from lmpy.data_wrangling.matrix.base import _MatrixDataWrangler


# .....................................................................................
class MatchTreeMatrixWrangler(_MatrixDataWrangler):
    """Subset and reorder a matrix to match a tree."""
    name = 'MatchTreeMatrixWrangler'
    version = '1.0'

    # .......................
    def __init__(
        self,
        tree,
        species_axis,
        **params
    ):
        """Constructor for MatchTreeMatrixWrangler class.

        Args:
            tree (TreeWrapper or str): A tree object or file path to a tree.
            species_axis (int): The matrix axis with species names to match.
            **params (dict): Keyword parameters to pass to _MatrixDataWrangler.
        """
        _MatrixDataWrangler.__init__(self, **params)

        if isinstance(tree, str):
            tree = TreeWrapper.from_filename(tree)
        self.tree = tree
        self.species_axis = species_axis
        self.report['changes'] = {}

    # .......................
    def wrangle_matrix(self, matrix):
        """Wrangle a matrix.

        Args:
            matrix (Matrix): A matrix to wrangle.

        Returns:
            Matrix: The matrix subset to tree tip names.
        """
        slices = []
        for axis in range(matrix.ndim):
            if axis == self.species_axis:
                axis_slice = []
                axis_headers = matrix.get_headers(axis=str(axis))

                for taxon in self.tree.taxon_namespace:
                    if taxon.label in axis_headers:
                        axis_slice.append(axis_headers.index(taxon.label))

                if str(axis) not in self.report['changes'].keys():
                    self.report['changes'][str(axis)] = {'purged': 0}
                self.report[
                    'changes'
                ][str(axis)]['purged'] += (len(axis_headers) - len(axis_slice))
            else:
                axis_slice = list(range(matrix.shape[axis]))

            slices.append(axis_slice)

        return matrix.slice(*slices)



"""Module containing occurrence data wranglers for filtering points."""
from lmpy.data_wrangling.occurrence.base import _OccurrenceDataWrangler


# .....................................................................................
class UniqueLocalitiesFilter(_OccurrenceDataWrangler):
    """Get an occurrence data wrangler for unique localities per species."""
    name = 'UniqueLocalitiesFilter'
    version = '1.0'

    # .......................
    def __init__(self, do_reset=True, **params):
        """Get an occurrence data wrangler unique localities for a group of points.

        Args:
            do_reset (bool): Reset the list of seen localities after each group.
            **params (dict): Extra parameters to be sent to the base class.
        """
        self.seen_localities = []
        # JSON may make boolean into a string so handle that
        if isinstance(do_reset, str):
            do_reset = do_reset.lower() != 'false'
        self.do_reset = bool(do_reset)
        _OccurrenceDataWrangler.__init__(self, **params)

    # .......................
    def _pass_condition(self, point):
        """Assess a point to see if it is unique spatially.

        Args:
            point (Point): A point object to assess.

        Returns:
            bool: An indication if the point is spatially unique.
        """
        test_val = (point.species_name, point.x, point.y)
        if test_val in self.seen_localities:
            return False
        self.seen_localities.append(test_val)
        return True

    # .......................
    def wrangle_points(self, points):
        """Wrangle points.

        Adds an optional reset step so the same filter can be used for many species.

        Args:
            points (list of Point): A list of points to wrangle.

        Returns:
            list of Point: A list of points with unique localities.
        """
        if self.do_reset:
            self.seen_localities = []
        return _OccurrenceDataWrangler.wrangle_points(self, points)



"""Module containing Occurrence Data Wrangler base class."""
from lmpy.data_wrangling.base import _DataWrangler


# .....................................................................................
class _OccurrenceDataWrangler(_DataWrangler):
    """Occurrence data wrangler base class."""
    name = '_OccurrenceDataWrangler'

    # .......................
    def __init__(
        self,
        store_attribute=None,
        pass_value=0,
        fail_value=1,
        **params
    ):
        """Constructor for _OcccurrenceDataWrangler base class.

        Args:
            store_attribute (str or None): If provided, add an assessment field to the
                point.
            pass_value (object): If the point passes the assessment, set the field to
                this value.
            fail_value (object): If the point fails the assessment, set the field to
                this value.
            **params (dict): A dictionary of keyword parameters.
        """
        _DataWrangler.__init__(self, **params)
        self.assessed = 0
        self.filtered = 0
        self.modified = 0

        self.store_attribute = store_attribute
        self.pass_value = pass_value
        self.fail_value = fail_value

    # .......................
    def _modify_point(self, point):
        """A function that modifies a Point object as needed.

        The base class version is a dummy function that returns the input point.

        Args:
            point (Point): A point object to  modify.

        Returns:
            Point, bool: Return the provided point and False indicating that it was not
                modified.
        """
        return point, False

    # .......................
    def _pass_condition(self, point):
        """A function that determines if a point passes some criteria.

        The base class version always passes.

        Args:
            point (Point): A point object to assess.

        Returns:
            bool: Indication if the point passes the test condition.
        """
        return True

    # .......................
    def get_report(self):
        """Get a report of the wrangler's functioning.

        Returns:
            dict: A dictionary of wrangler outputs.
        """
        self.report['assessed'] = self.assessed
        self.report['modified'] = self.modified
        self.report['filtered'] = self.filtered
        return self.report

    # .......................
    def report_point(self, filtered=False, modified=False):
        """Report the result of wrangling a point.

        Args:
            filtered (bool): Was the point filtered.
            modified (bool): Was the point modified.
        """
        self.assessed += 1
        self.modified += int(modified)
        self.filtered += int(filtered)

    # .......................
    def wrangle_points(self, points):
        """Wrangle occurrence `Point` objects.

        Args:
            points (list of Point): A list of points to wrangle.

        Returns:
            list of point: A list of wrangled occurrnece points.
        """
        wrangled_points = []
        for point in points:
            pt = self.wrangle_single_point(point)
            if pt is not None:
                wrangled_points.append(pt)
        return wrangled_points

    # .......................
    def wrangle_single_point(self, point):
        """Wrangle a single point.

        Args:
            point (Point): A point object to wrangle.

        Returns:
            Point or None: An assessed and / or modified point or None if filtered.
        """
        # Do any modifications necessary for the point
        mod_point, is_modified = self._modify_point(point)
        is_filtered = False
        val = self.pass_value

        # Check if the point passes the pass condition
        if not self._pass_condition(point):
            is_filtered = True
            val = self.fail_value

        # If we should just assess the point, set the attribute
        if self.store_attribute is not None:
            mod_point.set_attribute(self.store_attribute, val)
            is_modified = True
        elif is_filtered:
            # If it was filtered and we should remove, do so
            mod_point = None

        self.report_point(filtered=is_filtered, modified=is_modified)

        return mod_point



"""Tool for building shapegrid shapefiles."""
import argparse

from lmpy.data_preparation.build_grid import build_shapegrid


DESCRIPTION = '''\
Build a Shapegrid shapefile that is used to define the sites of a PAM or other \
multivariate Matrix for lmpy operations.'''


# .....................................................................................
def cli():
    """Command-line interface to build shapegrid."""
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        'shapegrid_filename',
        type=str,
        help='File location to write shapegrid shapefile.',
    )
    parser.add_argument('min_x', type=float, help='The minimum x value for the grid.')
    parser.add_argument('min_y', type=float, help='The minimum y value for the grid.')
    parser.add_argument('max_x', type=float, help='The maximum x value for the grid.')
    parser.add_argument('max_y', type=float, help='The maximum y value for the grid.')
    parser.add_argument(
        'cell_size',
        type=float,
        help='The resolution of each grid cell (in EPSG map units).',
    )
    parser.add_argument(
        'epsg', type=int, help='The EPSG code (map projection) for the shapegrid.'
    )
    args = parser.parse_args()
    cell_sides = 4  # Add this to parameters if we enable hexagons again
    build_shapegrid(
        args.shapegrid_filename,
        args.min_x,
        args.min_y,
        args.max_x,
        args.max_y,
        args.cell_size,
        args.epsg,
        cell_sides,
    )


# .....................................................................................
__all__ = ['build_shapegrid', 'cli']


# .....................................................................................
if __name__ == '__main__':  # pragma: no cover
    cli()



"""Module containing Matrix Data Wrangler base class."""
from lmpy.data_wrangling.base import _DataWrangler


# .....................................................................................
class _MatrixDataWrangler(_DataWrangler):
    """Matrix data wrangler base class."""
    name = '_MatrixDataWrangler'

    # .......................
    def __init__(self, **params):
        """Constructor for _MatrixDataWrangler base class.

        Args:
            **params (dict): Named parameters to pass to _DataWrangler base class.
        """
        _DataWrangler.__init__(self, **params)
        self.report['changes'] = {}

    # .......................
    def _report_slice(self, axis, idx, modified=False, purged=False):
        """Report what is done on a matrix slice.

        Args:
            axis (int): The axis of the interaction.
            idx (int): The index of the interaction.
            modified (bool): Was the slice modified.
            purged (bool): Was the slice purged.
        """
        if str(axis) not in self.report['changes']:
            self.report['changes'][str(axis)] = {
                'axis': axis,
                'modified': 0,
                'purged': 0
            }
        if modified:
            self.report['changes'][str(axis)]['modified'] += 1

        if purged:
            self.report['changes'][str(axis)]['purged'] += 1

    # .......................
    def wrangle_matrix(self, matrix):
        """Wrangle a matrix.

        Args:
            matrix (Matrix): A matrix to wrangle.

        Raises:
            NotImplementedError: This method is not implemeneted for the base class.
        """
        raise NotImplementedError('wrangle_matrix not implemented on base class.')

    # .......................
    def reorder(self, matrix, *axes_order):
        """Reorder / subset a matrix for a number of dimensions.

        Args:
            matrix (Matrix): A matrix to reorder.
            *axes_order (list of iterables): Iterables for each dimension of a matrix
                to reorder / subset.

        Returns:
            Matrix: A reordered and / or subset matrix.
        """
        return matrix.slice(*axes_order)



"""Module containing occurrence data wranglers for modifying point data."""
import numpy as np

from lmpy.data_wrangling.common.accepted_name_wrangler import (
    _AcceptedNameWrangler,
    resolve_names_gbif,
)
from lmpy.data_wrangling.matrix.base import _MatrixDataWrangler


# .....................................................................................
class AcceptedNameMatrixWrangler(_MatrixDataWrangler, _AcceptedNameWrangler):
    """Modifies matrix columns to update taxon names to the "accepted" names."""
    name = 'AcceptedNameMatrixWrangler'
    version = '1.0'

    # .......................
    def __init__(
        self,
        name_map=None,
        name_resolver=None,
        taxon_axis=1,
        purge_failures=True,
        **params
    ):
        """Constructor for AcceptedNameMatrixModifier class.

        Args:
            name_map (dict): A map of original name to accepted name.
            name_resolver (str or Method): If provided, use this method for getting new
                accepted names.  If set to 'gbif', use GBIF name resolution.
            taxon_axis (int): The axis with taxon headers.
            purge_failures (bool): Should failures be purged from the matrix.
            **params (dict): Keyword parameters to pass to _MatrixDataWrangler.
        """
        _MatrixDataWrangler.__init__(self, **params)
        if isinstance(name_resolver, str) and name_resolver.lower() == 'gbif':
            name_resolver = resolve_names_gbif
        _AcceptedNameWrangler.__init__(
            self,
            name_map=name_map,
            name_resolver=name_resolver
        )

        self.taxon_axis = taxon_axis
        self.purge_failures = purge_failures

    # .......................
    def wrangle_matrix(self, matrix):
        """Wrangle a matrix.

        Args:
            matrix (Matrix): A matrix to wrangle.

        Returns:
            Matrix: The matrix with accepted taxon names.
        """
        failures = []
        headers = matrix.get_headers(axis=self.taxon_axis)
        for i, hdr in enumerate(headers):
            acc_name = self.resolve_names([hdr])[hdr]
            if acc_name is None:
                failures.append(i)
            elif hdr != acc_name:
                matrix.headers[str(self.taxon_axis)][i] = acc_name
                # report
                self._report_slice(self.taxon_axis, i, modified=True)

        # Set defaults for purged and modified
        purged = False
        modified = True

        # Should we purge?
        if self.purge_failures:
            new_headers = matrix.get_headers(str(self.taxon_axis))
            matrix = np.delete(matrix, failures, axis=self.taxon_axis)
            # Go through failures and pop each, reverse list to start at biggest to
            #     maintain indices
            for i in sorted(failures, reverse=True):
                new_headers.pop(i)
            matrix.set_headers(str(self.taxon_axis), new_headers)
            # Change purged and modified since we are removing items
            purged = True
            modified = False

        for i in failures:
            self._report_slice(self.taxon_axis, i, modified=modified, purged=purged)

        return matrix



"""Script to encode layers into a Matrix."""
import argparse
import os

from lmpy.data_preparation.layer_encoder import LayerEncoder
from lmpy.tools._config_parser import _process_arguments


DESCRIPTION = 'Encode raster and / or vector layers into a site by layer matrix.'


# .....................................................................................
def cli():
    """Command line interface for layer encoding.

    Raises:
        ValueError: Raised if an unknown encoding method is provided or too many layer
            arguments.
    """
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument('--config_file', type=str, help='Path to configuration file.')
    parser.add_argument(
        '--encode_method',
        '-m',
        type=str,
        choices=['biogeo', 'presence_absence', 'largest_class', 'mean_value'],
        default='mean_value',
        help='The method to use to encode each layer.',
    )
    parser.add_argument(
        '--min_coverage',
        type=float,
        default=25,
        help='Minimum percentage of a cell that has to be covered to encode it.',
    )
    parser.add_argument(
        '--min_presence',
        type=float,
        default=10,
        help='Minimum value to be considered present when encoding presence absence.',
    )
    parser.add_argument(
        '--max_presence',
        type=float,
        default=127,
        help='The maximum value that should be considered present.',
    )
    parser.add_argument(
        'shapegrid_filename',
        type=str,
        help='File location of shapegrid to use for site geometries.',
    )
    parser.add_argument(
        'out_matrix_filename',
        type=str,
        help='File location to write the encoded matrix.',
    )
    # Layers to encode
    parser.add_argument(
        '--layer',
        '-l',
        nargs='*',
        action='append',
        help='File location of layer [ LABEL [ EVENT FIELD ]].',
    )

    args = _process_arguments(parser, config_arg='config_file')

    encoder = LayerEncoder(args.shapegrid_filename)

    for layer_args in args.layer:
        # if there are too many layer arguments, fail
        if len(layer_args) > 3:
            raise ValueError(
                'Too many layer arguments {}.'
                'Hint: Specify layer args last in command.'.format(layer_args)
            )
        # Process layer arguments, get filename for sure, try label and event field
        lyr_fn = layer_args[0]
        layer_event = None

        # Label
        if len(layer_args) > 1:
            lyr_label = layer_args[1]
            if len(layer_args) > 2:
                layer_event = layer_args[2]
        else:
            lyr_label = os.path.splitext(os.path.basename(lyr_fn))[0]

        if args.encode_method == 'biogeo':
            encoder.encode_biogeographic_hypothesis(
                lyr_fn, lyr_label, args.min_coverage, event_field=layer_event
            )
        elif args.encode_method == 'presence_absence':
            encoder.encode_presence_absence(
                lyr_fn,
                lyr_label,
                args.min_presence,
                args.max_presence,
                args.min_coverage,
                attribute_name=layer_event,
            )
        elif args.encode_method == 'largest_class':
            encoder.encode_largest_class(
                lyr_fn,
                lyr_label,
                args.min_coverage,
                attribute_name=layer_event,
            )
        elif args.encode_method == 'mean_value':
            encoder.encode_mean_value(
                lyr_fn, lyr_label, attribute_name=layer_event
            )
        else:
            raise ValueError('Unknown encoding method: {}'.format(args.encode_method))

    enc_mtx = encoder.get_encoded_matrix()
    enc_mtx.write(args.out_matrix_filename)


# .....................................................................................
__all__ = ['cli']


# .....................................................................................
if __name__ == '__main__':  # pragma: no cover
    cli()



"""Module containing functions for splitting occurrence data."""
import os

from lmpy.point import PointCsvWriter


DEFAULT_MAX_WRITERS = 100


# .....................................................................................
def get_writer_filename_func(base_dir):
    """Get a function that returns a filename from a writer key.

    Args:
        base_dir (str): A base directory for all writers.

    Returns:
        Method: A function that returns a filename when given a writer key.
    """
    # ..................
    def get_writer_filename_from_key(writer_key):
        """Get a writer filename from a writer key and create directories if needed.

        Args:
            writer_key (str or list): A writer key to use to generate a filename.

        Returns:
            str: A file path to use for the writer.
        """
        try:
            writer_fn = '{}.csv'.format(os.path.join(base_dir, writer_key))
        except TypeError:  # Tried to join iterable
            writer_fn = '{}.csv'.format(os.path.join(base_dir, *writer_key))
        os.makedirs(os.path.dirname(writer_fn), exist_ok=True)
        return writer_fn

    return get_writer_filename_from_key


# .....................................................................................
def get_writer_key_from_fields_func(*fields):
    """Get a function that returns a writer key from fields of a point.

    Args:
        *fields (list): A list of fields to use to determine the Point's writer key.

    Returns:
        Method: A function that takes a Point as an argument and returns a key.
    """
    key_fields = tuple(fields)

    # .......................
    def key_from_fields_func(point):
        """Get a wrangler key for a point.

        Args:
            point (Point): A point object to get a writer key for.

        Returns:
            Object: An object representing the key for the particular point.
        """
        writer_key = [point.get_attribute(fld) for fld in key_fields]
        if len(writer_key) == 1:
            return writer_key[0]
        return writer_key

    return key_from_fields_func


# .....................................................................................
class OccurrenceSplitter:
    """A tool for splitting occurrence data by some criteria for easier processing."""

    # .......................
    def __init__(
        self,
        writer_key_func,
        writer_filename_func,
        write_fields=None,
        max_writers=DEFAULT_MAX_WRITERS,
    ):
        """Constructor.

        Args:
            writer_key_func (Method): A function for determining a writer to use.  It
                should take a Point as input and return a dictionary key.
            writer_filename_func (Method): A function to determine the file location
                to write data for a particular writer.  It should take a dictionary
                key and return a string.
            write_fields (list): A list of fields to write for each writer.  If None,
                use all fields in the first output Point object.
            max_writers (int): The maximum number of open writers (files) at any given
                time.
        """
        self.get_writer_key = writer_key_func
        self.get_writer_filename = writer_filename_func
        self.writer_fields = write_fields
        self.max_writers = max_writers
        self.writers = {}

    # .......................
    def __enter__(self):
        """Context manager magic method.

        Returns:
            OccurrenceSplitter: This instance.
        """
        return self

    # .......................
    def __exit__(self, *args):
        """Context manager magic method on exit.

        Args:
            *args: Position arguments passed to the method.
        """
        self.flush_writers()

    # .......................
    def flush_writers(self):
        """Close all open occurrence writers."""
        for writer in self.writers.values():
            writer.close()
        self.writers = {}

    # .......................
    def close(self):
        """Close all open writers."""
        self.flush_writers()

    # .......................
    def open_writer(self, writer_key):
        """Open an occurrence writer for the provided key.

        Args:
            writer_key (object): Some key that can be used to determine a writer.
        """
        # Flush or close writers if needed
        if len(self.writers) >= self.max_writers:
            self.flush_writers
        # Open the new writer
        writer_fn = self.get_writer_filename(writer_key)
        if os.path.exists(writer_fn):
            self.writers[writer_key] = PointCsvWriter(
                writer_fn, self.writer_fields, mode='at', write_headers=False
            )
        else:
            self.writers[writer_key] = PointCsvWriter(
                writer_fn, self.writer_fields, mode='wt', write_headers=True
            )
        self.writers[writer_key].open()

    # .......................
    def process_reader(self, reader, wranglers):
        """Process an occurrence reader.

        Args:
            reader (PointCsvReader or PointDwcaReader): An occurrence reader instance.
            wranglers (list): A list of occurrence data wranglers.
        """
        reader.open()
        for points in reader:
            for wrangler in wranglers:
                if points:
                    points = wrangler.wrangle_points(points)
            if points:
                self.write_points(points)
        reader.close()

    # .......................
    def write_points(self, points):
        """Write points using the appropriate writer.

        Args:
            points (list): A list of point objects to write to file.
        """
        if points:
            writer_key = self.get_writer_key(points[0])
            if writer_key not in self.writers.keys():
                if self.writer_fields is None:
                    self.writer_fields = list(points[0].attributes.keys())
                self.open_writer(writer_key)
            self.writers[writer_key].write_points(points)


# .....................................................................................
__all__ = [
    'get_writer_key_from_fields_func',
    'get_writer_filename_func',
    'OccurrenceSplitter',
]



"""Spatial tools."""
# flake8: noqa
from . import geojsonify
from . import spatial_index
from .geojsonify import *
from .spatial_index import *

__all__ = []
__all__.extend(geojsonify.__all__)
__all__.extend(spatial_index.__all__)



"""Module containing occurrence data wranglers for filtering points."""
from lmpy.data_wrangling.occurrence.base import _OccurrenceDataWrangler


# .....................................................................................
class BoundingBoxFilter(_OccurrenceDataWrangler):
    """Get an occurrence data wrangler for filtering by bounding box."""
    name = 'BoundingBoxFilter'
    version = '1.0'

    # .......................
    def __init__(self, min_x, min_y, max_x, max_y, **params):
        """Get an occurrence data wrangler for filtering by bounding box.

        Args:
            min_x (numeric): The minimum 'x' value for the bounding box.
            min_y (numeric): The minimum 'y' value for the bounding box.
            max_x (numeric): The maximum 'x' value for the bounding box.
            max_y (numeric): The maximum 'y' value for the bounding box.
            **params (dict): Keyword parameters to pass to _OccurrenceDataWrangler.
        """
        self.min_x = min_x
        self.min_y = min_y
        self.max_x = max_x
        self.max_y = max_y
        _OccurrenceDataWrangler.__init__(self, **params)

    # .......................
    def _pass_condition(self, point):
        """Pass condition for a point.

        Args:
            point (Point): A point object to assess.

        Returns:
            bool: Indication if the point passed assessment.
        """
        return all(
            [self.min_x <= point.x <= self.max_x, self.min_y <= point.y <= self.max_y]
        )



"""Test script to process an iDigBio DWCA into something we can parse."""
import argparse
import io
import zipfile

import defusedxml.ElementTree as ET

from lmpy.point import PointDwcaReader


# DWCA Tag Constants
CORE_TAG = '{http://rs.tdwg.org/dwc/text/}core'
FIELD_TAG = '{http://rs.tdwg.org/dwc/text/}field'
FILES_TAG = '{http://rs.tdwg.org/dwc/text/}files'
ID_TAG = '{http://rs.tdwg.org/dwc/text/}id'
LOCATION_TAG = '{http://rs.tdwg.org/dwc/text/}location'
EXTENSION_TAG = '{http://rs.tdwg.org/dwc/text/}extension'

ROW_TYPE_ATT = 'rowType'
OCCURRENCE_ROW_TYPE = 'http://rs.tdwg.org/dwc/terms/Occurrence'

TERM_TYPE_ATT = 'term'
DELIMITED_TERMS = [
    'http://portal.idigbio.org/terms/flags',
    'http://portal.idigbio.org/terms/recordIds',
]
DELIMITED_BY_ATT = 'delimitedBy'


# .....................................................................................
def process_meta_xml(meta_xml_contents):
    """Process the contents of the meta.xml file.

    Args:
        meta_xml_contents (str): String containing XML metadata information for a DWCA.

    Returns:
        str: The processed metadata XML as a string.
    """
    meta_xml_root = ET.fromstring(meta_xml_contents)
    core_element = meta_xml_root.find(CORE_TAG)

    # If core element is missing (iDigBio) look in extensions
    if core_element is None:
        for extension_el in meta_xml_root.findall(EXTENSION_TAG):
            if (
                core_element is None
                and extension_el.attrib[ROW_TYPE_ATT] == OCCURRENCE_ROW_TYPE
            ):
                core_element = extension_el
                extension_el.tag = CORE_TAG

    # Add attribute for delimited fields
    for field_element in core_element.findall(FIELD_TAG):
        if field_element.attrib[TERM_TYPE_ATT] in DELIMITED_TERMS:
            field_element.attrib[DELIMITED_BY_ATT] = ';'

    return ET.tostring(meta_xml_root)


# .....................................................................................
def process_idb_dwca(in_zipfile, out_zipfile):
    """Process an idigbio zipfile.

    Args:
        in_zipfile (str): File path to an input DWCA zip file.
        out_zipfile (str): File path to write the output DWCA zip file.
    """
    # Open zip files
    dwca_in = zipfile.ZipFile(in_zipfile, mode='r')
    dwca_out = zipfile.ZipFile(out_zipfile, mode='w')

    # Read and prcoess meta.xml
    meta_xml_contents = io.TextIOWrapper(dwca_in.open('meta.xml')).read()
    new_meta_xml_contents = process_meta_xml(meta_xml_contents)
    dwca_out.writestr('meta.xml', new_meta_xml_contents)

    # Read meta.xml
    #   Correct element tag
    #   Add delimiter to fields?

    # Open occurrence.csv
    in_occ = io.TextIOWrapper(dwca_in.open('occurrence.csv'))

    i = 0
    trip_quote = '"' + '""'  # So python doesn't interpret as block comment
    quad_quote = '""' + '""'
    trip_quote_replace = '"' + "'"
    with open('temp_occ.csv', mode='wt') as temp_occ:
        for line in in_occ:
            i += 1
            # Remove double-double-quotes
            mod_line = (
                line.replace(quad_quote, '""')
                .replace(trip_quote, trip_quote_replace)
                .replace('""', "'")
            )

            quote_chunks = mod_line.split('"')
            write_line = ''

            inside = False

            for chunk in quote_chunks:
                if chunk.startswith('[') and chunk.endswith(']'):
                    write_line += (
                        chunk[1:-1].replace(',', ';').replace("'", '').replace(' ', '')
                    )
                elif chunk.startswith('{') and chunk.endswith('}'):  # JSON
                    write_line += '"' + chunk + '"'
                else:
                    # Quoted section
                    if inside:
                        write_line += '"' + chunk + '"'
                    else:
                        write_line += chunk
                inside = not inside

            # Process lists
            # while mod_line.find('"[') >= 0:
            #    pre_idx = mod_line.find('"[')
            #    post_idx = mod_line[pre_idx:].find(']"') + pre_idx

            #    pre_chunk = mod_line[:pre_idx]
            #    list_chunk = mod_line[pre_idx+2:post_idx].replace(',', ';')
            #        .replace("'", '').replace(' ', '')
            #    post_chunk = mod_line[post_idx+2:]

            #    mod_line = pre_chunk + list_chunk + post_chunk

            #    #mod_line = parts_0[0] + parts_1[0].replace(',', ';')
            #        .replace("'", '').replace(' ', '') + ']"'.join(parts_1[1:])

            # Process json sections (geopoints)

            if not write_line.endswith('\n'):
                write_line += '\n'
            temp_occ.write(write_line)

    dwca_out.write('temp_occ.csv', 'occurrence.csv')

    in_occ.close()

    # Close zip files
    dwca_in.close()
    dwca_out.close()


# .....................................................................................
def test_dwca(dwca_filename):
    """Test that a DWCA file can be processed.

    Args:
        dwca_filename (str): The file path of a DarwinCore Archive to test.

    Raises:
        ValueError: Raised if points returned is None.
    """
    with PointDwcaReader(
        dwca_filename, geopoint_term='geoPoint', x_term='lon', y_term='lat'
    ) as reader:
        for points in reader:
            if points is None:
                raise ValueError('Points is None')


# .....................................................................................
def main():
    """Main method for script."""
    parser = argparse.ArgumentParser()
    parser.add_argument('in_dwca', type=str, help='Input DWCA zip file.')
    parser.add_argument('out_dwca', type=str, help='Output processed DWCA zip file.')
    args = parser.parse_args()
    process_idb_dwca(args.in_dwca, args.out_dwca)
    test_dwca(args.out_dwca)


# .....................................................................................
if __name__ == '__main__':  # pragma: no cover
    main()



"""Tests the randomize swap method module."""
import numpy as np

from lmpy import Matrix
from lmpy.randomize.swap import swap_randomize, trial_swap
from copy import deepcopy


# .............................................................................
class Test_swap_randomize:
    """Test the swap randomize method."""

    # .....................................
    def _test_pam_randomization(self, pam):
        """Helper function to test randomization.

        Args:
            pam (Matrix): A presence-absence matrix to use for randomization testing.

        Raises:
            Exception: Raised if the PAM is not randomized.
        """
        rand = swap_randomize(deepcopy(pam), 10000)
        # Randomize until not the same as original
        count = 10
        while np.all(rand == pam) and count > 0:
            rand = swap_randomize(deepcopy(pam), 10000)
            count -= 1
        if count <= 0:
            raise Exception('Randomized PAM is same as original')

        # Test that marginal totals are the same
        assert np.all(np.sum(pam, axis=1) == np.sum(rand, axis=1))
        assert np.all(np.sum(pam, axis=0) == np.sum(rand, axis=0))

    # .....................................
    def test_empty_column(self):
        """Test with a known matrix that has an empty column."""
        pam = Matrix(
            np.array(
                [[0, 1, 0, 1, 0], [1, 1, 0, 0, 0], [0, 0, 0, 0, 1], [1, 1, 0, 1, 1]]
            )
        )
        self._test_pam_randomization(pam)

    # .....................................
    def test_empty_row(self):
        """Test with a known matrix that has an empty row."""
        pam = Matrix(
            np.array(
                [[0, 1, 0, 1, 0], [1, 1, 1, 0, 0], [0, 0, 0, 0, 0], [1, 1, 0, 1, 1]]
            )
        )
        self._test_pam_randomization(pam)

    # .....................................
    def test_full_column(self):
        """Test with a known matrix that has a full column."""
        pam = Matrix(
            np.array(
                [[0, 1, 1, 1, 0], [1, 1, 1, 0, 0], [0, 0, 1, 0, 1], [1, 1, 1, 1, 1]]
            )
        )
        self._test_pam_randomization(pam)

    # .....................................
    def test_full_row(self):
        """Test with a known matrix that has a full row."""
        pam = Matrix(
            np.array(
                [[0, 1, 0, 1, 0], [1, 1, 1, 1, 1], [0, 0, 1, 0, 1], [1, 1, 0, 1, 1]]
            )
        )
        self._test_pam_randomization(pam)

    # .....................................
    def test_known_matrix(self):
        """Test with a known matrix."""
        pam = Matrix(
            np.array(
                [[0, 1, 0, 1, 0], [1, 1, 1, 0, 0], [0, 0, 1, 0, 1], [1, 1, 0, 1, 1]]
            )
        )
        self._test_pam_randomization(pam)

    # .....................................
    def test_random_large_matrix(self):
        """Test with a randomly generated large matrix."""
        pam = Matrix(np.random.randint(0, 2, size=(1000, 10000)))
        self._test_pam_randomization(pam)

    # .....................................
    def test_random_matrix(self):
        """Test with a randomly generated matrix."""
        pam = Matrix(np.random.randint(0, 2, size=(10, 10)))
        self._test_pam_randomization(pam)


# .............................................................................
class Test_trial_swap_randomize:
    """Test the trial swap randomize method."""

    # .....................................
    def _test_pam_randomization(self, pam):
        """Helper function to test randomization.

        Args:
            pam (Matrix): A presence-absence matrix to use for randomization testing.

        Raises:
            Exception: Raised if the PAM is not randomized.
        """
        rand = trial_swap(deepcopy(pam))
        # Randomize until not the same as original
        count = 10
        while np.all(rand == pam) and count > 0:
            rand = trial_swap(deepcopy(pam))
            count -= 1
        if count <= 0:
            raise Exception('Randomized PAM is same as original')

        # Test that marginal totals are the same
        assert np.all(np.sum(pam, axis=1) == np.sum(rand, axis=1))
        assert np.all(np.sum(pam, axis=0) == np.sum(rand, axis=0))

    # .....................................
    def test_empty_column(self):
        """Test with a known matrix that has an empty column."""
        pam = Matrix(
            np.array(
                [[0, 1, 0, 1, 0], [1, 1, 0, 0, 0], [0, 0, 0, 0, 1], [1, 1, 0, 1, 1]]
            )
        )
        self._test_pam_randomization(pam)

    # .....................................
    def test_empty_row(self):
        """Test with a known matrix that has an empty row."""
        pam = Matrix(
            np.array(
                [[0, 1, 0, 1, 0], [1, 1, 1, 0, 0], [0, 0, 0, 0, 0], [1, 1, 0, 1, 1]]
            )
        )
        self._test_pam_randomization(pam)

    # .....................................
    def test_full_column(self):
        """Test with a known matrix that has a full column."""
        pam = Matrix(
            np.array(
                [[0, 1, 1, 1, 0], [1, 1, 1, 0, 0], [0, 0, 1, 0, 1], [1, 1, 1, 1, 1]]
            )
        )
        self._test_pam_randomization(pam)

    # .....................................
    def test_full_row(self):
        """Test with a known matrix that has a full row."""
        pam = Matrix(
            np.array(
                [[0, 1, 0, 1, 0], [1, 1, 1, 1, 1], [0, 0, 1, 0, 1], [1, 1, 0, 1, 1]]
            )
        )
        self._test_pam_randomization(pam)

    # .....................................
    def test_known_matrix(self):
        """Test with a known matrix."""
        pam = Matrix(
            np.array(
                [[0, 1, 0, 1, 0], [1, 1, 1, 0, 0], [0, 0, 1, 0, 1], [1, 1, 0, 1, 1]]
            )
        )
        self._test_pam_randomization(pam)

    # .....................................
    def test_random_large_matrix(self):
        """Test with a randomly generated large matrix."""
        pam = Matrix(np.random.randint(0, 2, size=(1000, 10000)))
        self._test_pam_randomization(pam)

    # .....................................
    def test_random_matrix(self):
        """Test with a randomly generated matrix."""
        pam = Matrix(np.random.randint(0, 2, size=(10, 10)))
        self._test_pam_randomization(pam)



"""Module containing data wranglers for subsetting a tree."""
from lmpy.data_wrangling.tree.base import _TreeDataWrangler


# .....................................................................................
class SubsetTreeWrangler(_TreeDataWrangler):
    """Subsets a tree."""
    name = 'SubsetTreeWrangler'
    version = '1.0'

    # .......................
    def __init__(
        self,
        keep_taxa,
        **params
    ):
        """Constructor for SubsetTreeWrangler class.

        Args:
            keep_taxa (list of str): A list of taxon names to keep.
            **params (dict): Keyword parameters to pass to _MatrixDataWrangler.
        """
        _TreeDataWrangler.__init__(self, **params)
        self.keep_taxon_names = keep_taxa

    # .......................
    def wrangle_tree(self, tree):
        """Wrangle a tree.

        Args:
            tree (TreeWrapper): A tree to wrangle.

        Returns:
            TreeWrapper: The subsetted tree.
        """
        original_taxa_count = len(tree.taxon_namespace)
        tree.retain_taxa_with_labels(self.keep_taxon_names)
        tree.purge_taxon_namespace()
        self._purged += (original_taxa_count - len(tree.taxon_namespace))
        return tree



"""Module containing occurrence data wranglers for filtering points."""
from osgeo import ogr
from lmpy.data_wrangling.occurrence.base import _OccurrenceDataWrangler


# .....................................................................................
class IntersectGeometriesFilter(_OccurrenceDataWrangler):
    """Get an occurrence data wrangler for filtering by intersect geometries."""
    name = 'IntersectGeometriesFilter'
    version = '1.0'

    # .......................
    def __init__(self, geometry_wkts, **params):
        """Get an occurrence data wrangler for filtering by intersecting geometries.

        Args:
            geometry_wkts (list of str): A list of WKT strings.
            **params (dict): Keyword parameters to pass to _OccurrenceDataWrangler.
        """
        _OccurrenceDataWrangler.__init__(self, **params)
        self.geometries = []
        for wkt in geometry_wkts:
            self.geometries.append(ogr.CreateGeometryFromWkt(wkt))

    # .......................
    def _pass_condition(self, point):
        """Assessment function for a point.

        Args:
            point (Point): A point object to assess.

        Returns:
            bool: An indication if a point passed assessment.
        """
        point_geometry = ogr.Geometry(ogr.wkbPoint)
        point_geometry.AddPoint(point.x, point.y)
        return any(
            [
                not geom.Intersection(point_geometry).IsEmpty()
                for geom in self.geometries
            ]
        )



"""Module containing a data wrangler base class for resolving taxon names."""
import json
import requests
import time
import urllib

from lmpy.data_wrangling.base import _DataWrangler


# .....................................................................................
def resolve_names_gbif(names, wait_time=.5):
    """Resolve names using GBIF's taxonomic name resolution service.

    Args:
        names (list of str): A list of name strings to resolve.
        wait_time (number): A number of seconds to wait after each request to avoid
            server ire.

    Returns:
        dict: Input names are keys and resolved name or None are values.
    """
    resolved_names = {}
    for name_str in names:
        # Get name
        other_filters = {'name': name_str.strip(), 'verbose': 'true'}
        url = 'http://api.gbif.org/v1/species/match?{}'.format(
            urllib.parse.urlencode(other_filters))
        response = requests.get(url).json()
        if response['status'].lower() in ('accepted', 'synonym'):
            resolved_names[name_str] = response['canonicalName']
        else:
            resolved_names[name_str] = None
        if wait_time is not None:
            time.sleep(wait_time)

    return resolved_names


# .....................................................................................
class _AcceptedNameWrangler(_DataWrangler):
    """Base class for accepted taxon name wranglers."""
    # .......................
    def __init__(self, name_map=None, name_resolver=None):
        """Constructor for the base accepted name wrangler.

        Args:
            name_map (dict or str or None): An existing name mapping.
            name_resolver (Method or None): If provided, this should be a function that
                takes a list of names as input and returns a dictionary of name
                mappings.  If omitted, resolving of new names will be skipped.
        """
        if name_map is not None:
            self._load_name_map(name_map)
        else:
            self.name_map = {}
        self._name_resolver = name_resolver

    # .......................
    def _load_name_map(self, name_map):
        """Attempt to load names from the name_map provided.

        Args:
            name_map (dict or str): A mapping dictionary or a filename with names.
        """
        if isinstance(name_map, dict):
            self.name_map = name_map
        else:
            self.name_map = {}
            try:
                # Try to load JSON names
                with open(name_map, mode='rt') as in_json:
                    self.name_map = json.load(in_json)
            except json.JSONDecodeError:  # Not a valid json file, try csv
                with open(name_map, mode='rt') as in_csv:
                    for line in in_csv:
                        in_name, out_name = line.strip().split(',')
                        self.name_map[in_name] = out_name

    # .......................
    def resolve_names(self, names):
        """Attempts to resolve a list of names.

        Args:
            names (list or str): A list of names to resolve.

        Returns:
            dict: A dictionary of input name keys and resolved name values.
        """
        if isinstance(names, str):
            names = [names]
        resolved_names = {}
        unmatched_names = []
        for name in names:
            if name in self.name_map.keys():
                resolved_names[name] = self.name_map[name]
            else:
                unmatched_names.append(name)
                resolved_names[name] = None

        # If we have a name resolver and names to resolve, do it
        if self._name_resolver is not None and len(unmatched_names) > 0:
            new_names = self._name_resolver(unmatched_names)
            # Update name map and return dictionary
            self.name_map.update(new_names)
            resolved_names.update(new_names)

        return resolved_names

    # .......................
    def write_map_to_file(self, filename, output_format, mode='wt'):
        """Write the name map to a file so it can be reused.

        Args:
            filename (str): A file location where the map should be written.
            output_format (str): The format to write the map, either 'csv' or 'json'.
            mode (str): How the file should be opened.
        """
        if output_format.lower() == 'json':
            with open(filename, mode=mode) as out_json:
                json.dump(self.name_map, out_json)
        else:
            with open(filename, mode=mode) as out_csv:
                out_csv.write('Name,Accepted\n')
                for in_name, out_name in self.name_map.items():
                    out_csv.write(f'{in_name},{out_name}\n')



"""Tools module."""
from . import build_shapegrid
from . import calculate_pam_stats
from . import clean_occurrences
from . import convert_csv_to_lmm
from . import convert_lmm_to_csv
from . import create_rare_species_model
from . import create_tree_matrix
from . import encode_layers
from . import mcpa_run
from . import randomize_pam
from . import split_occurrence_data
from . import wrangle_matrix
from . import wrangle_tree

__all__ = []
__all__.extend(build_shapegrid.__all__)
__all__.extend(calculate_pam_stats.__all__)
__all__.extend(clean_occurrences.__all__)
__all__.extend(convert_csv_to_lmm.__all__)
__all__.extend(convert_lmm_to_csv.__all__)
__all__.extend(create_rare_species_model.__all__)
__all__.extend(create_tree_matrix.__all__)
__all__.extend(encode_layers.__all__)
__all__.extend(mcpa_run.__all__)
__all__.extend(randomize_pam.__all__)
__all__.extend(split_occurrence_data.__all__)
__all__.extend(wrangle_matrix.__all__)
__all__.extend(wrangle_tree.__all__)



"""Module containing occurrence data wranglers for purging empty matrix slices."""
import numpy as np

from lmpy.data_wrangling.matrix.base import _MatrixDataWrangler


# .....................................................................................
class PurgeEmptySlicesWrangler(_MatrixDataWrangler):
    """Removes empty slices (rows, columns, other) from a matrix."""
    name = 'PurgeEmptySlicesWrangler'
    version = '1.0'

    # .......................
    def __init__(
        self,
        purge_axes=None,
        **params
    ):
        """Constructor for PurgeEmptySlicesWrangler class.

        Args:
            purge_axes (None, int, or list of int): Matrix axes to be processed.
            **params (dict): Keyword parameters to pass to _MatrixDataWrangler.
        """
        _MatrixDataWrangler.__init__(self, **params)
        # Make purge_axes a list if it is not None
        if isinstance(purge_axes, int):
            purge_axes = [purge_axes]
        self.purge_axes = purge_axes
        self.report['changes'] = {}

    # .......................
    def wrangle_matrix(self, matrix):
        """Wrangle a matrix.

        Args:
            matrix (Matrix): A matrix to wrangle.

        Returns:
            Matrix: The matrix with empty slices purged.
        """
        slices = []
        for axis in range(matrix.ndim):
            if self.purge_axes is None or axis in self.purge_axes:
                sum_axes = list(range(matrix.ndim))
                sum_axes.remove(axis)
                axis_slice = np.where(matrix.sum(axis=tuple(sum_axes)) != 0)[0]

                if str(axis) not in self.report['changes'].keys():
                    self.report['changes'][str(axis)] = {'purged': 0}
                self.report['changes'][str(axis)]['purged'] += len(
                    np.where(matrix.sum(axis=tuple(sum_axes)) == 0)[0]
                )
            else:
                axis_slice = list(range(matrix.shape[axis]))
            slices.append(axis_slice)

        return matrix.slice(*slices)



# -*- coding: utf-8 -*-
"""Setup module for lmpy."""
from setuptools import setup, find_packages
import versioneer


with open('README.md') as f:
    readme = f.read()

with open('LICENSE') as f:
    license = f.read()

setup(
    name='specify-lmpy',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Package of commonly used and shared Lifemapper objects and tools',
    long_description=readme,
    author='Specify Systems Lifemapper Team',
    author_email='cjgrady@ku.edu',
    url='https://github.com/specifysystems/lmpy',
    license=license,
    packages=find_packages(exclude=('tests', 'docs')),
    install_requires=[
        'defusedxml',
        'dendropy>=4.0.0',
        'numpy>=1.11.0',
        'gdal',
        'requests',
        'rtree',
    ],
)



"""Module containing tools for creating GeoJSON from matrices."""
import json

from osgeo import ogr


# .....................................................................................
def _process_omit_values(omit_values, matrix_dtype):
    """Process the incoming omit values to ensure they match matrix values type.

    Args:
        omit_values (list or None): A list of incoming omit values.
        matrix_dtype (type): The data type of the matrix values

    Returns:
        list: A list of values converted to the same type as the compare matrix.
    """
    if omit_values is None:
        return []
    return [matrix_dtype(val) for val in omit_values]


# .....................................................................................
def _get_geojson_geometry_func(resolution=None):
    """Get a function that will generate GeoJSON geometry sections for an x, y pair.

    Args:
        resolution (Numeric or None): If None, use point geometries, else polygons.

    Returns:
        Method: A function taking x, y pairs and returning dict geometry reference.
    """
    if resolution is not None:
        half_res = resolution / 2

    # .......................
    def _make_point_geojson_geometry(x, y):
        return {
            'type': 'Point',
            'coordinates': [x, y]
        }

    # .......................
    def _make_polygon_geojson_geometry(x, y):
        return {
            'type': 'Polygon',
            'coordinates': [
                [
                    [x - half_res, y - half_res],
                    [x + half_res, y - half_res],
                    [x + half_res, y + half_res],
                    [x - half_res, y + half_res],
                    [x - half_res, y - half_res],
                ]
            ]
        }

    if resolution is None:
        return _make_point_geojson_geometry
    return _make_polygon_geojson_geometry


# .....................................................................................
def geojsonify_matrix(matrix, resolution=None, omit_values=None):
    """Creates GeoJSON for a matrix and matching shapefile.

    Args:
        matrix (Matrix): A (spatial) matrix to create GeoJSON for.
        resolution (Numeric): The size of the grid cells in decimal degrees
        omit_values (list): Omit properties when their value is in this list.

    Returns:
        dict: A GeoJSON compatible dictionary.
    """
    omit_values = _process_omit_values(omit_values, matrix.dtype.type)
    ret = {'type': 'FeatureCollection'}
    features = []

    make_geometry_func = _get_geojson_geometry_func(resolution=resolution)

    row_headers = matrix.get_row_headers()
    column_headers = matrix.get_column_headers()

    column_enum = [(j, str(k)) for j, k in enumerate(column_headers)]

    for i, (site_id, x, y) in enumerate(row_headers):
        ft_json = dict(type='Feature', geometry=make_geometry_func(x, y))
        ft_json['id'] = site_id
        ft_json['properties'] = {
            k: matrix[i, j].item() for j, k in column_enum
            if matrix[i, j].item() not in omit_values
        }
        if len(ft_json['properties'].keys()) > 0:
            features.append(ft_json)

    ret['features'] = features
    return ret


# .....................................................................................
def geojsonify_matrix_with_shapefile(matrix, shapegrid_filename, omit_values=None):
    """Creates GeoJSON for a matrix and matching shapefile.

    Args:
        matrix (Matrix): A (spatial) matrix to create GeoJSON for.
        shapegrid_filename (str): A file path to a shapefile matching the matrix.
        omit_values (list): Omit properties when their value is in this list.

    Returns:
        dict: A GeoJSON compatible dictionary.
    """
    omit_values = _process_omit_values(omit_values, matrix.dtype.type)
    ret = {'type': 'FeatureCollection'}
    features = []

    column_headers = matrix.get_column_headers()

    column_enum = [(j, str(k)) for j, k in enumerate(column_headers)]

    shapegrid_dataset = ogr.Open(shapegrid_filename)
    shapegrid_layer = shapegrid_dataset.GetLayer()

    i = 0
    feat = shapegrid_layer.GetNextFeature()
    while feat is not None:
        ft_json = json.loads(feat.ExportToJson())
        ft_json['properties'] = {
            k: matrix[i, j].item() for j, k in column_enum
            if matrix[i, j].item() not in omit_values
        }
        if len(ft_json['properties'].keys()) > 0:
            features.append(ft_json)
        i += 1
        feat = shapegrid_layer.GetNextFeature()

    ret['features'] = features
    shapegrid_dataset = shapegrid_layer = None
    return ret


# .....................................................................................
__all__ = ['geojsonify_matrix', 'geojsonify_matrix_with_shapefile']
